install.packages('expm')
library(expm)

################### Functions for MDP construction #################

randomDistribution = function(n){  
  # This function generates a random discrete probability distribution
  # on n items
  # Input: n, a positive integer
  # Ouput: randomVector, a vector of length n with random components
  #         between 0 and 1 that sum to 1
  A = runif(n);
  randomVector = A/sum(A);
  return(randomVector)
}

makeProbArray = function(n,m){
  # This function repeatedly calls randomDistribution() to construct
  # an array with all probabilities necessary for a decision process
  # Input: n, number of states 
  #        m, number of actions
  # Output: probArray, a n x n x m array where probArray[i,j,k] is
  #           the probability of transitioning to state j when starting
  #           in state i and action k is used
  probArray = array(dim=c(n,n,m))
  for(i in 1:m){
    for(j in 1:n){
      probArray[j,,i] = randomDistribution(n)
    }
  }
  return(probArray)
}

inducedToOriginal = function(k,n,m){
  # This function accepts an integer, interpreted as an index for a state
  # in the induced process, and returns the state from the original process
  # as well as the history matrix.
  # Input: k, the index for a state in the induced process
  #        n, the number of states in the original process
  #        a, the number of actions in the original process
  # Output: returnList, a list with two named elements:
  #            originalState, representing the state from the original process
  #            historyMatrix, the history matrix
  historyMatrix = matrix(ncol=m,nrow=n)
  originalState = ((k-1) %/% (2^m-1)^n) + 1
  historyRemainder = ((k-1) %% (2^m-1)^n)
  for (i in n:1){
    # This loop extracts the history for each state at a time.
    historyQuotient = (historyRemainder %/% (2^m-1)^(i-1))
    historyVector = as.numeric(as.logical(intToBits(historyQuotient)))[1:m]
    if (sum(historyVector)==m){
      historyVector = rep(0,m)
    }
    historyMatrix[i,] = historyVector
    historyRemainder = (historyRemainder %% (2^m-1)^(i-1))
  }
  returnList = list("originalState"=originalState, "historyMatrix" = historyMatrix)
  return(returnList)
}

originalToInduced = function(originalState,historyMatrix){
  # This function accepts a state from the original process and a history
  # matrix, and returns the index for the corresponding state in the
  # induced process.
  # Inputs: originalState, a state from the original process
  #         historyMatrix, a history matrix
  # Outputs: k, the index of the corresponding state in the induced process
  m = ncol(historyMatrix)
  n = nrow(historyMatrix)
  historyContribution = 0
  for (i in 1:n){
    for (j in 1:m){
      if (historyMatrix[i,j]){
        historyContribution = historyContribution + 2^(j-1)*(2^m-1)^(i-1)
      }
    }
  }
  k = 1 + (originalState-1)*(2^m-1)^n+historyContribution
  return(k)
}

makeInducedChain = function(probArray){
  # This function constructs the probability transition matrix 
  # for the Markov chain induced by balanced wandering.
  # Input: probArray, an array containing probabilities for the original process
  # Output: transitionMatrix, the transition matrix for the induced chain
  n = dim(probArray)[1] # Recover number of states in original chain
  m = dim(probArray)[3] # Recover number of actions
  numStates = n*(2^m-1)^n
  transitionMatrix = matrix(rep(0,numStates^2),nrow=numStates,ncol=numStates)
  for (i in 1:numStates){
    # Extract the original state and history matrix
    iAsOriginal = inducedToOriginal(i,n,m)
    originalState = iAsOriginal$originalState
    historyMatrix = iAsOriginal$historyMatrix
    
    # The history for the original state will change once an action is selected, so extract it
    historyVector = historyMatrix[originalState,]
    
    # Find out how many actions are available to be selected from
    numEligibleActions=sum(!historyVector)
    for (j in which(historyVector==0)){          # For every eligible action
      nextHistoryVector = historyVector          # nextHistory is almost the same...
      nextHistoryVector[j]=1                     # except that the action used is no longer eligible
      if(sum(nextHistoryVector)==m){
        nextHistoryVector = rep(0,m)             # If all actions have been used, reset the history.
      }
      # Reconstruct the nextHistory matrix
      nextHistoryMatrix = historyMatrix
      nextHistoryMatrix[originalState,] = nextHistoryVector
      for (k in 1:n){                      # Assuming action j is used, loop over destination states
        transitionProbability = probArray[originalState,k,j]/numEligibleActions
        transitionMatrix[i,originalToInduced(k,nextHistoryMatrix)] = transitionProbability
      }
    }
  }
  rownames(transitionMatrix)=paste("row",1:numStates, sep="")
  colnames(transitionMatrix)=paste("col",1:numStates, sep="")
  return(transitionMatrix)
}

################  Functions for calculating expected hitting times #########

constructQR = function(transitionMatrix,n,m){
  # This function extracts Q, the submatrix of the transition matrix
  # for the induced process containing transitions between transient
  # states.
  # Input: transitionMatrix, the probability transition matrix for the induced process
  #        n, the number of states in the original process
  #        m, the number of actions in the original process          
  # Output: returnList, a list with two named elements:
  #           Q, the matrix with transition probabilities between transient states
  #           R, the matrix with transition probabilities from 
  #             transient to absorbing states.
  absorbingStates=c(n)
  numStates = n*(2^m-1)^n
  transientStates = setdiff(1:numStates,absorbingStates)
  transientInducedStates = vector("numeric",length=0)
  for (i in 1:numStates){
    originalState = inducedToOriginal(i,n,m)$originalState 
    if (is.element(originalState,transientStates)){
      transientInducedStates = append(transientInducedStates,i)
    }
  }
  absorbingInducedStates = setdiff(1:numStates,transientInducedStates)
  Q = transitionMatrix[transientInducedStates,transientInducedStates]
  R = transitionMatrix[transientInducedStates,absorbingInducedStates]
  returnList = list("Q"=Q,"R"=R)
  return(returnList)
}

expectedHittingTimeRAS = function(probArray){
  # This function calculates the expected value of the time until an 
  # absorbing state is reached for the first time under random action selection
  # Input: probArray, an array containing probabilities for the original process
  #        absorbingStates, a vector of states intended to be absorbing
  # Output: expectedHittingTime, E[tau_R]
  n=dim(probArray)[1]
  m=dim(probArray)[3]
  transitionMatrixRAS = matrix(,nrow=n,ncol=n)
  absorbingStates=c(n)
  for(i in 1:n){
    for(j in 1:n){
      transitionMatrixRAS[i,j] = mean(probArray[i,j,])
    }
  }
  transientStates = setdiff(1:n,absorbingStates)
  Q = transitionMatrixRAS[transientStates,transientStates]
  R = transitionMatrixRAS[transientStates,absorbingStates]
  ones = rep(1,length(transientStates))
  expectedVector = solve((diag(length(transientStates))-Q),ones)
  expectedHittingTime = expectedVector[1]
  return(expectedHittingTime) 
}

expectedHittingTimeBW = function(probArray){
  # This function calculates the expected value of the time until an 
  # absorbing state is reached for the first time under balanced wandering
  # Input: probArray, an array containing probabilities for the original process
  #        absorbingStates, a vector of states intended to be absorbing
  # Output: expectedHittingTime, E[tau_B]
  n=dim(probArray)[1]
  m=dim(probArray)[3]
  absorbingStates=c(n)
  transitionMatrix = makeInducedChain(probArray)
  Q = constructQR(transitionMatrix,n,m)$Q
  R = constructQR(transitionMatrix,n,m)$R
  numInducedTransientStates = nrow(Q)
  ones = rep(1,numInducedTransientStates)
  expectedVector = solve((diag(numInducedTransientStates)-Q),ones)
  expectedHittingTime = as.numeric(expectedVector[1])
  return(expectedHittingTime)
}

experiment = function(iterations, n, m){
  # This function repeatedly generates a random MDP, calculates the hitting
  # time under each strategy, and stores the differences.
  # Input: iterations, the number of random MDPs to generate
  #        n, the number of states
  #        m, the number of actions
  # Output: data, a vector with length 'iterations' of differences in hitting times
  data = vector(length=iterations)
  for(k in 1: iterations){
    probArray = makeProbArray(n,m)
    data[k] = expectedHittingTimeBW(probArray)-expectedHittingTimeRAS(probArray)
    print(k)
  }
  return(data)
}


################### Examples of how to use function ##################

n=3                                 # Set the number of states
m=3                                 # Set the number of actions
probArray = makeProbArray(n,m)      # Create a random MDP
expectedHittingTimeRAS(probArray)   # Get hitting time with-replacement
expectedHittingTimeBW(probArray)    # Get hitting time without-replacement



################### Functions for MDPs with dead ends #####################

randomDistributionInaccessible = function(n){
  # This function generates a random discrete probability distribution
  # on n items, but allows some transition probabilities to be zero
  # Input: n, a positive integer
  # Ouput: randomVector, a vector of length n with random components
  #         between 0 and 1 that sum to 1
  numZeros = sample(1:(n-1),1)
  indicesZeros = sample(1:n,numZeros)
  indicesNonZeros = setdiff(1:n,indicesZeros)
  A = runif(n-numZeros);
  nonZeroEntries = A/sum(A);
  randomVector = rep(0,n)
  k=1 # An index for which non zero entries have been used so far
  for (i in 1:n){
    if(is.element(i,indicesNonZeros)){
      randomVector[i]=nonZeroEntries[k]
      k = k+1
    }
  }
  return(randomVector)
}
makeProbArrayInaccessible = function(n,m){
  # This function repeatedly calls randomDistributionInaccessible() 
  # to construct an array with all probabilities necessary for a 
  # decision process.  It is possible, but not guaranteed, the MDP
  # will have a dead end.
  # Input: n, number of states 
  #        m, number of actions
  # Output: probArray, a n x n x m array where probArray[i,j,k] is
  #           the probability of transitioning to state j when starting
  #           in state i and action k is used
  
  probArray = array(dim=c(n,n,m))
  goalDistribution = rep(0,n)
  goalDistribution[n] = 1
  for(i in 1:m){
    for(j in 1:(n-1)){
      probArray[j,,i] = randomDistributionInaccessible(n)
    }
    probArray[n,,i] = goalDistribution
  }
  return(probArray)
}
findProbArrayDeadEnds = function(n,m){
  # This function repeatedly calls findProbArrayInaccessible() until
  # it generates a process with a dead end.
  stopflag=FALSE
  while(!stopflag){
    probArray=makeProbArrayInaccessible(n,m)
    transitionMatrixRAS = matrix(,nrow=n,ncol=n)
    for(i in 1:n){
      for(j in 1:n){
        transitionMatrixRAS[i,j] = mean(probArray[i,j,])
      }
    }
    B = transitionMatrixRAS%^%n
    if((B[1,n]>0) && is.element(0,B[2:(n-1),4])){
      stopflag=TRUE
    }
  }
  return(probArray)
}
alterProbArray = function(probArray){
  # This functions accepts a probability array representing a decision
  # process with dead ends, and routes the dead ends through a return state
  # back to the initial state.
  # Input: probArray, an array with transition probabilities for a decision
  #        process with n states and m actions
  #        m, number of actions
  # Output: probArrayDeadEnds, an (n+1) x (n+1) x m array, where the
  #         additional state is the return state
  
  # Find the states that are dead ends
  transitionMatrixRAS = matrix(,nrow=n,ncol=n)
  for(i in 1:n){
    for(j in 1:n){
      transitionMatrixRAS[i,j] = mean(probArray[i,j,])
    }
  }
  B=transitionMatrixRAS%^%n
  deadEnds=vector(,length=1)
  k=1
  for (i in 2:(n-1)){
    if(B[i,n]==0){
      deadEnds[k]=i
      k=k+1
    }
  }
  # Modify the transition matrix to account for dead ends
  # Introduce a return state at index n, and move goal to n+1
  probArrayDeadEnds = array(dim=c(n+1,n+1,m))
  for(i in 1:m){ # Loop over actions
    for(j in 1:(n-1)){
      for(k in 1:(n-1)){
        probArrayDeadEnds[j,k,i]=probArray[j,k,i]
      }
      probArrayDeadEnds[j,n,i]=0
      probArrayDeadEnds[j,n+1,i]=probArray[j,n,i]
    }
    probArrayDeadEnds[n,,i] = c(1,rep(0,n))
    probArrayDeadEnds[n+1,,i] = c(rep(0,n),1)
    for(d in deadEnds){
      probArrayDeadEnds[d,,i] = c(rep(0,n-1),1,0)
    }
  }
  return(probArrayDeadEnds)
}

################### The MDP from Figure 2 ###############################

probArray = array(0,dim=c(n+1,n+1,m))
probArray[1,1,1]=.62
probArray[1,3,1]=.16
probArray[1,5,1]=.22
probArray[2,4,1]=1
probArray[3,2,1]=.02
probArray[3,3,1]=.29
probArray[3,5,1]=.69
probArray[4,1,1]=1
probArray[5,5,1]=1
probArray[1,1,2]=.23
probArray[1,2,2]=.36
probArray[1,5,2]=.41
probArray[2,4,2]=1
probArray[3,2,2]=.41
probArray[3,3,2]=.02
probArray[3,5,2]=.57
probArray[4,1,2]=1
probArray[5,5,2]=1

a=expectedHittingTimeRAS(probArray)
b=expectedHittingTimeBW(probArray)
a-b  # This difference being negative shows BW is worse than RAS.
